redo-ifchange \
    ../config \
    *.texi \
    cfg/*.texi \
    cmd/*.texi \
    integration/*.texi \
    pedro.txt \
    pkt/*.texi \
    sp.plantuml.txt \
    usecases.ru/*.texi \
    usecases/*.texi
. ../config
${MAKEINFO:-makeinfo} \
    -D "VERSION `../version`" \
    $MAKEINFO_OPTS \
    --set-customization-variable SECTION_NAME_IN_TITLE=1 \
    --set-customization-variable TREE_TRANSFORMATIONS=complete_tree_nodes_menus \
    --set-customization-variable ASCII_PUNCTUATION=1 \
    --output $3 index.texi
